@@ -5,8 +5,8 @@
     sub execute {
         my $self = shift;
         my ( $controller, $c ) = @_;
-        $self->prepare_soap_helper($c);
-        $self->NEXT::execute(@_);
+        $self->prepare_soap_helper($controller,$c);
+        $self->next::method(@_);
     }
 };
 
